import logging
import os
import time
from pathlib import Path

from huggingface_hub import snapshot_download
from huggingface_hub.utils import RepositoryNotFoundError, RevisionNotFoundError

logger = logging.getLogger(__name__)


def download_and_link_snapshot(
    hf_repo_id: str, hf_token: str, local_dir: Path, max_retries: int, retry_delay: int
):
    """Download a snapshot using the local HF cache and symlink it to local dir"""
    logger.info("Downloading model snapshot...")

    for attempt in range(max_retries):
        try:
            snapshot_path = snapshot_download(
                hf_repo_id,
                token=hf_token,
                etag_timeout=60 * 10,
                resume_download=True,
            )
            logger.info(f"Model snapshot downloaded successfully to {snapshot_path}.")

            for item in Path(snapshot_path).iterdir():
                d = local_dir / item.name
                if d.is_symlink():
                    d.unlink()
                elif d.is_dir():
                    d.rmdir()
                elif d.exists():
                    d.unlink()

                if item.is_dir():
                    os.symlink(item, d, target_is_directory=True)
                else:
                    os.symlink(item, d)
            break
        except (ConnectionError, TimeoutError) as e:
            if attempt < max_retries - 1:
                logger.warning(f"Connection error: {e}. Retrying in {retry_delay}s...")
                time.sleep(retry_delay)
                retry_delay *= 2  # exponential backoff
            else:
                print(
                    f"Failed to download after {max_retries} attempts. Please check your internet connection."
                )
                raise
        except (RepositoryNotFoundError, RevisionNotFoundError) as e:
            logger.error(
                "Please check the repository ID and ensure you have the correct permissions."
            )
            raise
        except Exception as e:
            logger.error("An unexpected error occurred.")
            raise



from briton.proto import FinishReason


async def collect_text(async_text_iter, eos_token: str) -> tuple[str, int]:
    full_text = ""
    num_completion_tokens = 0
    async for delta in async_text_iter:
        num_completion_tokens += len(delta.output_ids)
        full_text += delta.output_text
    # HACK(@bdubayah): To avoid refactoring, just add the eos token if the finish reason should be "stop"
    if delta.finish_reason == FinishReason.END_ID or delta.finish_reason == FinishReason.STOP_WORDS:
        full_text += eos_token
    return full_text, num_completion_tokens



from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, List, Optional, Union

from fastapi import HTTPException
from openai.types.chat import ChatCompletion
from openai.types.completion import Completion

from briton.async_util import map_generator, tap_generator, try_advance_generator
from briton.backend.utils import collect_text
from briton.constants import BRITON_DEFAULT_MAX_TOKENS, EXECUTOR_API_ENABLED
from briton.data_structures import optional_if
from briton.error_handling import grpc_error_handling
from briton.lora_cache import LoraNotFoundException
from briton.openai import create_completion, create_completion_chunks
from briton.proto import (
    FinishReason,
    InferenceAnswerPart,
    InferenceRequest,
    LookaheadDecodingConfig,
)
from briton.schema import ModelInput


async def model_input_to_briton_request(
    request_id: int,
    model_input: ModelInput,
    input_ids: List[int],
    tokenizer_eos_token_id: Optional[int],
    tokenizer_pad_token_id: Optional[int],
    add_schema_to_cache: Callable[[Dict[str, Any]], Awaitable[str]],
    resolve_lora: Optional[Callable[[str], Optional[int]]],
    default_max_tokens: Optional[int],
    max_seq_len: int,
) -> InferenceRequest:
    """
    Convert ModelInput to Briton inference request.

    :param model_input: model input
    :return: InferRequest
    """
    request = InferenceRequest(
        request_id=request_id,
        input_ids=input_ids,
    )
    num_input_ids = len(input_ids)

    # end_id, pad_id
    end_id = model_input.end_id or tokenizer_eos_token_id
    if end_id is not None:
        request.end_id = end_id
    pad_id = model_input.pad_id or tokenizer_pad_token_id
    if pad_id is not None:
        request.pad_id = pad_id

    schema = model_input.output_json_schema
    if schema is not None and not EXECUTOR_API_ENABLED:
        request.output_schema_hash = await add_schema_to_cache(schema)
        force_tools = model_input.force_tools
        if force_tools is not None:
            request.force_tools = force_tools

    # guided decoding
    if model_input.guided_decoding_params is not None and EXECUTOR_API_ENABLED:
        request.guided_decoding_params.CopyFrom(model_input.guided_decoding_params)

    if model_input.model is not None and len(model_input.model) > 0 and resolve_lora is not None:
        try:
            lora_task_id = resolve_lora(model_input.model)
        except LoraNotFoundException:
            raise HTTPException(status_code=404, detail=f"Model {model_input.model} not found.")
        if lora_task_id is not None:
            request.lora_task_id = lora_task_id

    # max tokens
    request.request_output_len = calc_request_output_len(
        model_input.max_tokens,
        num_input_ids,
        default_max_tokens,
        max_seq_len,
    )

    # beam width
    if model_input.beam_width is not None:
        request.beam_width = model_input.beam_width

    # repetition penalty
    if model_input.repetition_penalty is not None:
        request.repetition_penalty = model_input.repetition_penalty

    # presence_penalty
    if model_input.presence_penalty is not None:
        request.presence_penalty = model_input.presence_penalty

    # length penalty
    if model_input.length_penalty is not None:
        request.len_penalty = model_input.length_penalty

    # temperature
    if model_input.temperature is not None:
        request.temperature = model_input.temperature

    # length_penalty
    if model_input.length_penalty is not None:
        request.len_penalty = model_input.length_penalty

    # frequency_penalty
    if model_input.frequency_penalty is not None:
        # TODO(pankaj) Implement this in Briton
        # request.frequency_penalty = model_input.frequency_penalty
        pass

    # top_k
    if model_input.top_k is not None:
        request.runtime_top_k = model_input.top_k

    # top_p
    if model_input.top_p is not None:
        request.runtime_top_p = model_input.top_p

    # random seed
    # Note that model_input always has a seed. If seed is not supplied in input
    # then a random value is picked.
    request.random_seed = model_input.seed

    # stop words
    if model_input.stop is not None:
        request.stop_words.extend(model_input.stop)

    # bad words
    if model_input.bad_words is not None:
        request.bad_words.extend(model_input.bad_words)

    if model_input.lookahead_decoding_config is not None:
        request.lookahead_decoding_config.CopyFrom(
            LookaheadDecodingConfig(
                window_size=model_input.lookahead_decoding_config.window_size,
                ngram_size=model_input.lookahead_decoding_config.ngram_size,
                verification_set_size=model_input.lookahead_decoding_config.verification_set_size,
            )
        )

    return request


def calc_request_output_len(
    model_input_max_tokens: Optional[int],
    num_input_ids: int,
    default_max_tokens: Optional[int],
    max_seq_len: int,
) -> int:
    if num_input_ids > max_seq_len:
        raise HTTPException(
            status_code=400,
            detail=f"Number of input tokens {num_input_ids} exceeds max sequence length {max_seq_len}",
        )
    model_max_allowed_tokens = max_seq_len - num_input_ids
    max_tokens = model_input_max_tokens or default_max_tokens or BRITON_DEFAULT_MAX_TOKENS
    return min(model_max_allowed_tokens, max_tokens)


async def openai_spec_response(
    resp_iters: List[AsyncGenerator[InferenceAnswerPart, None]],
    request_id: str,
    num_input_ids: int,
    streaming: bool,
    eos_token: str,
    tool_call_token: Optional[str],
    tool_call_id_fn: Optional[Callable[[], str]],
    model_name: str,
    include_stream_usage: bool,
    stop_words: Optional[List[str]],
    skip_special_tokens: Optional[List[str]],
    is_chat_completion: bool,
) -> Union[AsyncGenerator[str, None], ChatCompletion | Completion]:
    with grpc_error_handling():
        if streaming:
            num_completion_tokens = 0

            def get_num_completion_tokens():
                nonlocal num_completion_tokens
                return num_completion_tokens

            def set_num_completion_tokens(item: InferenceAnswerPart):
                nonlocal num_completion_tokens
                num_completion_tokens += len(item.output_ids)

            def transform(item: InferenceAnswerPart) -> str:
                output_text = item.output_text
                # HACK(@bdubayah): To avoid refactoring, just add the eos token if the finish reason should be "stop"
                if (
                    item.finish_reason == FinishReason.END_ID
                    or item.finish_reason == FinishReason.STOP_WORDS
                ):
                    output_text += eos_token
                return output_text

            sequences = [
                map_generator(
                    tap_generator(
                        await try_advance_generator(resp_iter), set_num_completion_tokens
                    ),
                    transform,
                )
                for resp_iter in resp_iters
            ]

            return create_completion_chunks(
                req_id=request_id,
                model=model_name,
                sequences=sequences,
                eos_token=eos_token,
                tool_token=tool_call_token,
                tool_call_id_fn=tool_call_id_fn,
                prompt_tokens=optional_if(num_input_ids, include_stream_usage),
                completion_tokens_fn=optional_if(get_num_completion_tokens, include_stream_usage),
                stop_words=stop_words,
                skip_special_tokens=skip_special_tokens,
                is_chat_completion=is_chat_completion,
            )
        else:
            total_num_completion_tokens = 0
            sequences = []
            for resp_iter in resp_iters:
                generation_text, num_completion_tokens = await collect_text(resp_iter, eos_token)
                sequences.append(generation_text)
                total_num_completion_tokens += num_completion_tokens
            return create_completion(
                req_id=request_id,
                model=model_name,
                sequences=sequences,
                eos_token=eos_token,
                tool_token=tool_call_token,
                tool_call_id_fn=tool_call_id_fn,
                prompt_tokens=num_input_ids,
                completion_tokens=total_num_completion_tokens,
                stop_words=stop_words,
                skip_special_tokens=skip_special_tokens,
                is_chat_completion=is_chat_completion,
            )



from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, List, Optional, Union

from briton.backend.backend_types import RequestDetails
from briton.backend.briton_request import model_input_to_briton_request
from briton.backend.default_backend import DefaultBackend
from briton.backend.utils import collect_text
from briton.data_structures import or_true
from briton.error_handling import grpc_error_handling
from briton.openai import remove_suffix_from_text
from briton.proto import InferenceAnswerPart
from briton.schema import ModelInput


class LegacyBackend(DefaultBackend):
    def __init__(self):
        super().__init__()

    # Overrides parent
    async def accepts_request(self, model_input: ModelInput) -> Optional[RequestDetails]:
        if self._config is None or self._tokenizer is None:
            return None

        if self._config.is_openai_compatible:
            return None

        input_ids = model_input.input_ids(self._tokenizer)
        return RequestDetails(input_ids=input_ids)

    async def infer(
        self,
        model_input: ModelInput,
        is_cancelled: Callable[[], Awaitable[bool]],  # TODO(pankaj) Wire up request cancellation
        add_schema_to_cache: Callable[[Dict[str, Any]], Awaitable[str]],
        resolve_lora: Optional[Callable[[str], Optional[int]]],
        request_details: RequestDetails,
    ) -> Union[AsyncGenerator[str, None], str]:
        if (
            self._config is None
            or self._tokenizer is None
            or self._briton_stub is None
            or self._generate_request_id is None
        ):
            raise ValueError("Model is not loaded.")
        input_ids = request_details.input_ids
        if input_ids is None:
            raise ValueError("Input ids are None.")

        request_id = self._generate_request_id()
        briton_request = await model_input_to_briton_request(
            request_id=request_id,
            model_input=model_input,
            input_ids=input_ids,
            tokenizer_eos_token_id=self._tokenizer.eos_token_id,
            tokenizer_pad_token_id=self._tokenizer.pad_token_id,
            add_schema_to_cache=add_schema_to_cache,
            resolve_lora=resolve_lora,
            default_max_tokens=self._config.default_max_tokens,
            max_seq_len=self._config.max_seq_len,
        )

        resp_iter = self._briton_stub.Infer(briton_request)

        return await self._plaintext_response(
            resp_iter=resp_iter,
            streaming=or_true(model_input.stream),
            eos_token=self._tokenizer.eos_token,
            tool_call_token=self._config.tool_call_token,
            stop_words=model_input.stop,
            skip_special_tokens=model_input.skip_special_tokens,
        )

    @staticmethod
    async def _plaintext_response(
        resp_iter: AsyncGenerator[InferenceAnswerPart, None],
        streaming: bool,
        eos_token: str,
        tool_call_token: Optional[str],
        stop_words: Optional[List[str]],
        skip_special_tokens: Optional[List[str]],
    ) -> Union[AsyncGenerator[str, None], str]:
        with grpc_error_handling():
            if streaming:
                # Advance the iterator to get the first chunk, to allow any
                # validation error to be thrown. Any errors will be handled
                # gy grpc_error_handling and converted to HTTPException.
                # `first_chunk` is not discarded, it is used below.
                async for first_chunk in resp_iter:
                    break

                async def generate_processed_after_first_chunk():
                    output_text = first_chunk.output_text
                    # Try to remove tool call token from first chunk only
                    if tool_call_token is not None:
                        output_text = output_text.removeprefix(tool_call_token)
                    yield remove_suffix_from_text(
                        text=output_text,
                        eos_token=eos_token,
                        stop_words=stop_words,
                        skip_special_tokens=skip_special_tokens,
                    )

                    async for chunk in resp_iter:
                        output_text = chunk.output_text
                        yield remove_suffix_from_text(
                            text=output_text,
                            eos_token=eos_token,
                            stop_words=stop_words,
                            skip_special_tokens=skip_special_tokens,
                        )

                return generate_processed_after_first_chunk()
            else:
                full_text, _ = await collect_text(resp_iter, eos_token)
                full_text = remove_suffix_from_text(
                    text=full_text,
                    eos_token=eos_token,
                    stop_words=stop_words,
                    skip_special_tokens=skip_special_tokens,
                )
                if tool_call_token is not None:
                    full_text = full_text.removeprefix(tool_call_token)
                return full_text



# TODO(pankaj) Pick this from truss constants, once a small lib is extracted out of truss.
import os
import string

OPENAI_NON_COMPATIBLE_TAG = "non-openai-compatible"
DEFAULT_BRITON_PORT = 50051
DEFAULT_TP_COUNT = 1

# Use a directory that can be picked up by baseten-fs, if enabled
FSM_CACHE_DIR = "/cache/model/fsm_cache"

# Directory where huggingface config.json files are uploaded by engine-builder.
# identical to engine-builder config.json
TOKENIZATION_DIR = "tokenization"

DEFAULT_MAX_FSM_WORKERS = 10

TOOL_CALL_IDS = {
    "llama": 128010,
    "mistral": 5,
    "palmyra": 151657,
    "qwen": 151657,
}

TOOL_CALL_TOKENS = {
    "llama": "<|python_tag|>",
    "mistral": "[TOOL_CALLS]",
    "palmyra": "<tool_call>",
    "qwen": "<tool_call>",
}

# vLLM uses the format `chatcmpl-tool-` followed by a length 32
# alphanumeric string: chatcmpl-tool-ace101101c7149f2b0ef11d5ef6bf694
#
# Note that Llama and its variants don't internally use the tool call id,
# so this id could be anything.
#
# Mistral's chat template expects a length 9 alphanumeric string, so we can't
# use the vLLM tool call id style.
USE_VLLM_TOOL_CALL_ID_STYLE = {
    "llama": True,
    "mistral": False,
    "palmyra": True,
    "qwen": True,
}


MODEL_INPUT_TO_BRITON_FIELD = {
    "max_tokens": "request_output_len",
    "beam_width": "beam_width",
    "repetition_penalty": "repetition_penalty",
    "presence_penalty": "presence_penalty",
    "temperature": "temperature",
    "length_penalty": "len_penalty",
    "end_id": "end_id",
    "pad_id": "pad_id",
    "runtime_top_k": "runtime_top_k",
    "runtime_top_p": "runtime_top_p",
    "random_seed": "random_seed",
    "stop_words": "stop_words",
    "bad_words": "bad_words",
}

# TODO(bdubayah): Don't hardcode this
LOCAL_PREDICT_ENDPOINT = "http://localhost:8080/v1/models/model:predict"

UINT32_MAX = 2**32 - 1

ALPHANUMERIC_CHARS = string.ascii_letters + string.digits

BRITON_DEFAULT_MAX_TOKENS = 50

TRT_CONFIG_FILENAME = "config.json"

ENABLE_EXECUTOR_API = "ENABLE_EXECUTOR_API"
EXECUTOR_API_ENABLED = (
    os.environ.get(ENABLE_EXECUTOR_API, "false").lower() == "true"
    or os.environ.get(ENABLE_EXECUTOR_API, "false").lower() == "1"
    or os.environ.get(ENABLE_EXECUTOR_API, "false").lower() == "yes"
)



from pathlib import Path
from typing import List


def is_local_path(path: str):
    return Path(path).exists()


def list_files(dir: Path) -> List[str]:
    """List files that are immediate children of the directory."""
    return [f.name for f in dir.iterdir() if f.is_file()]


def list_dirs(dir: Path) -> List[str]:
    return [f.name for f in dir.iterdir() if f.is_dir()]


def safe_mkdir(dir: Path) -> None:
    """Creates a directory, making parents if necessary and passing if the directory already exists."""
    if not dir.exists():
        dir.mkdir(parents=True, exist_ok=True)



import json
import logging
from pathlib import Path
from typing import Iterable, List, Optional, cast

from google.protobuf import text_format
from tokenizers import AddedToken

from briton.proto import AddedToken as AddedTokenProto
from briton.proto import (
    AddedTokens,
    BritonConfig,
    LookaheadDecodingConfig,
    XGrammarConfig,
)
from briton.trtllm_config import TrussTRTLLMBatchSchedulerPolicy


def trtllm_config_check(config):
    if "trt_llm" not in config:
        raise ValueError("trt_llm config is required for this model")


def batch_scheduler_policy_to_int(
    policy: TrussTRTLLMBatchSchedulerPolicy, logger: logging.Logger
) -> int:
    if policy == TrussTRTLLMBatchSchedulerPolicy.MAX_UTILIZATION:
        return 0
    elif policy == TrussTRTLLMBatchSchedulerPolicy.GUARANTEED_NO_EVICT:
        return 1
    else:
        logger.warning(f"Unknown batch scheduler policy: {policy}. Using GUARANTEED_NO_EVICT.")
        return 1


def create_briton_config_pbtxt(
    engine_path: str,
    hf_tokenizer: str,
    fsm_cache_dir: str,
    kv_cache_free_gpu_mem_fraction: float,
    port: int,
    added_tokens: list,
    max_num_tokens: Optional[int],
    config_pbtxt_path: Path,
    kv_cache_host_memory_bytes: Optional[int] = None,
    enable_kv_cache_reuse: bool = True,
    enable_chunked_context: bool = False,
    batch_scheduler_policy: Optional[int] = None,
    lookahead_decoding_config: Optional[LookaheadDecodingConfig] = None,
    runtime_max_batch_size: Optional[int] = None,
    xgrammar_config: Optional[XGrammarConfig] = None,
):
    briton_config = BritonConfig(
        engine_path=engine_path,
        hf_tokenizer=hf_tokenizer,
        kv_cache_free_gpu_mem_fraction=kv_cache_free_gpu_mem_fraction,
        kv_cache_host_memory_bytes=kv_cache_host_memory_bytes,
        enable_kv_cache_reuse=enable_kv_cache_reuse,
        enable_chunked_context=enable_chunked_context,
        port=port,
        fsm_cache_dir=fsm_cache_dir,
        max_num_tokens=max_num_tokens,
        batch_scheduler_policy=batch_scheduler_policy,
        added_tokens=AddedTokens(
            tokens=[
                AddedTokenProto(
                    content=token.content,
                    single_word=token.single_word,
                    lstrip=token.lstrip,
                    rstrip=token.rstrip,
                    normalized=token.normalized,
                    special=token.special,
                )
                for token in cast(Iterable[AddedToken], added_tokens)
            ]
        ),
        lookahead_decoding_config=lookahead_decoding_config,
        max_batch_size=runtime_max_batch_size,
        xgrammar_config=xgrammar_config,
    )
    with open(config_pbtxt_path, "w") as f:
        f.write(text_format.MessageToString(briton_config))


def engine_supports_chunked_context(engine_config_path: Path) -> bool:
    if not engine_config_path.exists():
        return False
    with engine_config_path.open() as f:
        engine_config = json.load(f)
        if "build_config" in engine_config and "plugin_config" in engine_config["build_config"]:
            plugin_config = engine_config["build_config"]["plugin_config"]
            context_fmha = plugin_config["context_fmha"]
            paged_kv_cache = plugin_config["paged_kv_cache"]
            use_paged_context_fmha = plugin_config["use_paged_context_fmha"]
            return context_fmha and paged_kv_cache and use_paged_context_fmha
    return False



SEED_OFFSET = 1021
UPPER_SEED_LIMIT = 2**63 - 1
LOWER_SEED_LIMIT = -(2**63)


def generate_next_random_seed(prev_seed) -> int:
    """Generate next random seed based on previous seed.

    Cycle through the seed range, if the next seed exceeds the upper limit.
    """
    next_seed = prev_seed + SEED_OFFSET
    if next_seed > UPPER_SEED_LIMIT:
        overflow = next_seed - UPPER_SEED_LIMIT
        next_seed = LOWER_SEED_LIMIT + overflow
    return next_seed



import logging
import time
from pathlib import Path
from typing import Dict, Optional

import grpc
import numpy as np

from briton.constants import DEFAULT_BRITON_PORT
from briton.fs import list_dirs, safe_mkdir
from briton.proto import BritonStub, DataType, InferenceRequest, Tensor

LORA_CONFIG_FILENAME = "model.lora_config.npy"
LORA_WEIGHTS_FILENAME = "model.lora_weights.npy"

NP_TO_BRITON_DTYPE = {
    np.int8: DataType.DT_INT8,
    np.uint8: DataType.DT_UINT8,
    np.int32: DataType.DT_INT32,
    np.int64: DataType.DT_INT64,
    np.float16: DataType.DT_FLOAT16,
    np.float32: DataType.DT_FLOAT32,
    np.bool_: DataType.DT_BOOL,
}

logger = logging.getLogger(__name__)


class LoraNotFoundException(Exception):
    pass


class LoraInvalidException(Exception):
    pass


class LoraCache:
    def __init__(self, base_model_name: str, loras_dir: Path):
        logger.info("Lora is enabled.")

        self._base_model_name = base_model_name
        self._loras_dir = loras_dir
        safe_mkdir(self._loras_dir)
        self._loaded_loras: Dict[str, int] = {}

        channel = grpc.insecure_channel(f"localhost:{DEFAULT_BRITON_PORT}")
        stub = BritonStub(channel)
        for i, lora_name in enumerate(list_dirs(self._loras_dir)):
            lora_task_id = i + 1
            self._load_lora(lora_name=lora_name, lora_task_id=lora_task_id, stub=stub)
        channel.close()

    def resolve_lora(self, lora_name: str) -> Optional[int]:
        if lora_name == self._base_model_name:
            return None
        if lora_name not in self._loaded_loras:
            raise LoraNotFoundException
        return self._loaded_loras[lora_name]

    def _load_lora(self, lora_name: str, lora_task_id: int, stub: BritonStub) -> None:
        logger.info(f"Loading LoRA {lora_name}...")
        start = time.time()

        if lora_name == self._base_model_name:
            raise LoraInvalidException

        lora_config = self._load_tensor(self._loras_dir / lora_name / LORA_CONFIG_FILENAME)
        lora_weights = self._load_tensor(self._loras_dir / lora_name / LORA_WEIGHTS_FILENAME)

        request = InferenceRequest(
            request_id=1,
            input_text=" ",
            request_output_len=1,
        )
        request.lora_task_id = lora_task_id
        request.lora_config.CopyFrom(lora_config)
        request.lora_weights.CopyFrom(lora_weights)

        for _ in stub.Infer(request):
            pass

        self._loaded_loras[lora_name] = lora_task_id

        stop = time.time()
        logger.info(f"Loaded LoRA {lora_name} in {stop - start}s!")

    @staticmethod
    def _load_tensor(np_tensor_path: Path) -> Tensor:
        if not np_tensor_path.exists():
            raise LoraInvalidException
        data = np.load(np_tensor_path, allow_pickle=True)
        tensor = Tensor()
        tensor.shape.dim.extend(data.shape)
        tensor.data = data.tobytes()
        tensor.dtype = NP_TO_BRITON_DTYPE[data.dtype.type]
        return tensor



import socket


def is_port_available(port, host="localhost"):
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            # Try to connect to the specified host and port
            s.bind((host, port))
            return True
    except OSError:
        # Port is not available
        return False



from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, Optional, Union

from openai.types.chat import ChatCompletion

from briton.backend.backend_types import (
    InferBackend,
    LazyLoadParams,
    LoadParams,
    RequestDetails,
)
from briton.backend.briton_request import (
    model_input_to_briton_request,
    openai_spec_response,
)
from briton.data_structures import or_default, or_false
from briton.openai import generate_tool_call_id, generate_tool_call_id_vllm
from briton.proto import InferenceRequest
from briton.schema import ModelInput


class DefaultBackend(InferBackend):
    """Regular OpenAI spec support backend."""

    def __init__(self):
        self._tokenizer = None
        self._briton_stub = None
        self._generate_request_id = None
        self._config = None

    def load(self, load_params: LoadParams) -> None:
        self._tokenizer = load_params.tokenizer
        self._generate_request_id = load_params.generate_request_id
        self._config = load_params.config

    async def lazy_load(self, lazy_load_params: LazyLoadParams) -> None:
        self._briton_stub = lazy_load_params.briton_stub

    async def accepts_request(self, model_input: ModelInput) -> Optional[RequestDetails]:
        if self._config is None or self._tokenizer is None:
            return None

        if not self._config.is_openai_compatible:
            return None

        input_ids = model_input.input_ids(self._tokenizer)
        return RequestDetails(input_ids=input_ids)

    @staticmethod
    def _tool_call_id_fn(
        use_vllm_tool_call_id_style: Optional[bool],
    ) -> Optional[Callable[[], str]]:
        if or_false(use_vllm_tool_call_id_style):
            return generate_tool_call_id_vllm
        else:
            return generate_tool_call_id

    async def infer(
        self,
        model_input: ModelInput,
        is_cancelled: Callable[[], Awaitable[bool]],  # TODO(pankaj) Wire up request cancellation
        add_schema_to_cache: Callable[[Dict[str, Any]], Awaitable[str]],
        resolve_lora: Optional[Callable[[str], Optional[int]]],
        request_details: RequestDetails,
    ) -> Union[AsyncGenerator[str, None], ChatCompletion]:
        if (
            self._config is None
            or self._tokenizer is None
            or self._briton_stub is None
            or self._generate_request_id is None
        ):
            raise ValueError("Model is not loaded.")

        input_ids = request_details.input_ids
        if input_ids is None:
            raise ValueError("Input ids are None.")

        request_id = self._generate_request_id()
        briton_request = await model_input_to_briton_request(
            request_id=request_id,
            model_input=model_input,
            input_ids=input_ids,
            tokenizer_eos_token_id=self._tokenizer.eos_token_id,
            tokenizer_pad_token_id=self._tokenizer.pad_token_id,
            add_schema_to_cache=add_schema_to_cache,
            resolve_lora=resolve_lora,
            default_max_tokens=self._config.default_max_tokens,
            max_seq_len=self._config.max_seq_len,
        )

        # TODO(bdubayah): replace with num_return_sequences in Briton
        # This is a workaround to send n requests, which would be equivalent to
        # sampling n return sequences in the same request
        resp_iters = [self._briton_stub.Infer(briton_request)]
        if model_input.n is not None and model_input.n > 1:
            for i in range(1, model_input.n):
                new_request = InferenceRequest()
                new_request.CopyFrom(briton_request)
                new_request.request_id = self._generate_request_id()
                new_request.random_seed = new_request.random_seed + 1
                resp_iters.append(self._briton_stub.Infer(new_request))

        return await openai_spec_response(
            resp_iters=resp_iters,
            request_id=str(request_id),
            num_input_ids=len(input_ids),
            streaming=or_false(model_input.stream),
            eos_token=self._tokenizer.eos_token,
            tool_call_token=self._config.tool_call_token,
            tool_call_id_fn=self._tool_call_id_fn(self._config.use_vllm_tool_call_id_style),
            model_name=or_default(model_input.model, ""),
            include_stream_usage=model_input.include_stream_usage,
            stop_words=model_input.stop,
            skip_special_tokens=model_input.skip_special_tokens,
            is_chat_completion=model_input._is_chat_completion,
        )



from .briton_pb2 import (
    AddedToken,
    AddedTokens,
    Batch,
    BritonConfig,
    InferenceRequest,
    InferenceAnswerPart,
    StatesToTokens,
    TokenToNextState,
    Tensor,
    DataType,
    FinishReason,
    GuidedDecodingParams,
    LookaheadDecodingConfig,
    XGrammarConfig,
)
from .briton_pb2_grpc import BritonStub


__all__ = [
    "BritonStub",
    "StatesToTokens",
    "TokenToNextState",
    "InferenceRequest",
    "InferenceAnswerPart",
    "Batch",
    "BritonConfig",
    "AddedToken",
    "AddedTokens",
    "Tensor",
    "DataType",
    "FinishReason",
    "GuidedDecodingParams",
    "LookaheadDecodingConfig",
    "XGrammarConfig",
]



import asyncio
from typing import AsyncGenerator, Awaitable, Callable, List, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def get_all_items(queue: asyncio.Queue):
    """Get all items from the queue without waiting."""
    items = []
    try:
        while True:
            item = queue.get_nowait()
            items.append(item)
    except asyncio.QueueEmpty:
        # Once the queue is empty, get_nowait raises QueueEmpty, and we stop
        pass
    return items


def set_event_loop_if_not_exist():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:  # 'RuntimeError' will be raised if there is no running loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def retry_predicate(
    pred: Callable[[], Awaitable[bool]],
    max_retries=5,
    delay_secs=2,
) -> Callable[[], Awaitable[bool]]:
    async def wrapper():
        for attempt in range(max_retries):
            try:
                if await pred():
                    return True
            except Exception:
                pass
            if attempt < max_retries - 1:
                await asyncio.sleep(delay_secs)
        return False

    return wrapper


def predicate_with_timeout(
    pred: Callable[[], Awaitable[bool]],
    timeout_secs: int,
) -> Callable[[], Awaitable[bool]]:
    async def wrapper():
        try:
            return await asyncio.wait_for(pred(), timeout=timeout_secs)
        except Exception:
            return False

    return wrapper


async def interleave_generators(*generators: AsyncGenerator[T, None]) -> AsyncGenerator[T, None]:
    """Interleaves multiple generators."""
    queue = asyncio.Queue()
    active = len(generators)

    async def feed_queue(generator: AsyncGenerator[T, None]):
        nonlocal active
        try:
            async for item in generator:
                await queue.put(item)
        except Exception as e:
            await queue.put(e)
        finally:
            active -= 1

    try:
        tasks = [asyncio.create_task(feed_queue(generator)) for generator in generators]

        while active > 0 or not queue.empty():
            item_or_exception = await queue.get()
            if isinstance(item_or_exception, Exception):
                raise item_or_exception
            yield item_or_exception
    finally:
        pending_tasks = [t for t in tasks if not t.done()]
        for task in pending_tasks:
            task.cancel()
        if pending_tasks:
            await asyncio.gather(*pending_tasks, return_exceptions=True)


async def try_advance_generator(generator: AsyncGenerator[T, None]) -> AsyncGenerator[T, None]:
    """Advances a generator once and returns the original generator with all items.

    Useful if an error might be thrown on the first iteration that needs to be caught.
    """
    it = aiter(generator)
    try:
        first = await anext(it)

        async def gen():
            yield first
            async for x in it:
                yield x

        return gen()
    except StopAsyncIteration:
        empty_list: List[T] = []

        async def gen():
            for x in empty_list:
                yield x

        return gen()


async def tap_generator(
    generator: AsyncGenerator[T, None], fn: Callable[[T], None]
) -> AsyncGenerator[T, None]:
    """Applies a function to each element of a generator without modifying the sequence."""
    async for item in generator:
        fn(item)
        yield item


async def map_generator(
    generator: AsyncGenerator[T, None], fn: Callable[[T], U]
) -> AsyncGenerator[U, None]:
    """Transforms each element of a generator."""
    async for item in generator:
        yield fn(item)



import json
import logging
import random
import time
from json import JSONDecodeError
from typing import Any, AsyncGenerator, Callable, List, Literal, Optional

import openai.types.chat.chat_completion as chat_completion
import openai.types.chat.chat_completion_chunk as chat_completion_chunk
from fastapi import HTTPException
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from openai.types.completion import Completion
from openai.types.completion_choice import CompletionChoice
from openai.types.completion_usage import CompletionUsage

from briton.async_util import interleave_generators
from briton.constants import ALPHANUMERIC_CHARS

logger = logging.getLogger(__name__)


def _load_content_json(content: str) -> Any:
    """Safely load the content json from the input text."""
    try:
        return json.loads(content)
    except JSONDecodeError:
        raise HTTPException(status_code=400, detail="Tool call was cut off by max_tokens.")


def generate_tool_call_id() -> str:
    return "".join(random.choices(ALPHANUMERIC_CHARS, k=9))


def generate_tool_call_id_vllm() -> str:
    return "chatcmpl-tool-" + "".join(random.choices(ALPHANUMERIC_CHARS, k=32))


def _create_tool_calls(
    content: str,
    tool_call_id_fn: Callable[[], str],
) -> List[ChatCompletionMessageToolCall]:
    content_json = _load_content_json(content)
    tool_calls = []
    for briton_fn in content_json:
        fn = Function(name=briton_fn["name"], arguments=json.dumps(briton_fn["parameters"]))
        tool_call = ChatCompletionMessageToolCall(
            id=tool_call_id_fn(), function=fn, type="function"
        )
        tool_calls.append(tool_call)
    return tool_calls


def _finish_reason_from_text(
    text: str, eos_token: Optional[str] = None, stop_words: Optional[List[str]] = None
) -> Literal["stop", "length"]:
    if eos_token and text.endswith(eos_token):
        return "stop"
    if stop_words and text.endswith(tuple(stop_words)):
        return "stop"
    return "length"


def remove_suffix_from_text(
    text: str,
    eos_token: Optional[str] = None,
    stop_words: Optional[List[str]] = None,
    skip_special_tokens: Optional[List[str]] = None,
) -> str:
    if eos_token and text.endswith(eos_token):
        return text.removesuffix(eos_token)
    if stop_words:
        for stop_word in stop_words:
            if text.endswith(stop_word):
                return text.removesuffix(stop_word)
    # HACK (bdubayah): this could end up being very expensive.
    if skip_special_tokens:
        for special_token in skip_special_tokens:
            text = text.replace(special_token, "")
    return text


def _create_choice(
    index: int,
    text: str,
    eos_token: Optional[str],
    tool_token: Optional[str],
    tool_call_id_fn: Optional[Callable[[], str]],
    stop_words: Optional[List[str]],
    skip_special_tokens: Optional[List[str]],
    is_chat_completion: bool,
) -> chat_completion.Choice:
    finish_reason = _finish_reason_from_text(text, eos_token, stop_words)
    content = remove_suffix_from_text(text, eos_token, stop_words, skip_special_tokens)
    if is_chat_completion:
        tool_calls = None
        if (
            tool_token is not None
            and tool_call_id_fn is not None
            and content.startswith(tool_token)
        ):
            finish_reason = "tool_calls"
            content = content.removeprefix(tool_token)
            tool_calls = _create_tool_calls(content, tool_call_id_fn)
            content = None
        message = ChatCompletionMessage(content=content, role="assistant", tool_calls=tool_calls)
        return chat_completion.Choice(finish_reason=finish_reason, index=index, message=message)
    else:
        if content is None:
            content = ""
        return CompletionChoice(finish_reason=finish_reason, index=index, text=content)


def create_completion(
    req_id: str,
    model: str,
    sequences: List[str],
    eos_token: Optional[str] = None,
    tool_token: Optional[str] = None,
    tool_call_id_fn: Optional[Callable[[], str]] = None,
    prompt_tokens: Optional[int] = None,
    completion_tokens: Optional[int] = None,
    stop_words: Optional[List[str]] = None,
    skip_special_tokens: Optional[List[str]] = None,
    is_chat_completion: bool = False,
) -> ChatCompletion:
    created = int(time.time())
    choices = []
    for i, text in enumerate(sequences):
        choice = _create_choice(
            index=i,
            text=text,
            eos_token=eos_token,
            tool_token=tool_token,
            tool_call_id_fn=tool_call_id_fn,
            stop_words=stop_words,
            skip_special_tokens=skip_special_tokens,
            is_chat_completion=is_chat_completion,
        )
        choices.append(choice)
    usage = None
    if prompt_tokens is not None and completion_tokens is not None:
        usage = CompletionUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )
    if is_chat_completion:
        return ChatCompletion(
            id=req_id,
            choices=choices,
            created=created,
            model=model,
            object="chat.completion",
            usage=usage,
        )
    else:
        return Completion(
            id=req_id,
            choices=choices,
            created=created,
            model=model,
            object="text_completion",
            usage=usage,
        )


def _make_sse_chunk(chunk: ChatCompletionChunk | Completion) -> str:
    return f"data: {chunk.model_dump_json()}\n\n"


async def _chunk_args(args_str: str) -> AsyncGenerator[str, None]:
    chunk = ""
    for char in args_str:
        chunk += char
        if char == " ":
            yield chunk
            chunk = ""
    if chunk:
        yield chunk


async def _create_tool_call_deltas(
    content: str,
    tool_call_id_fn: Callable[[], str],
) -> AsyncGenerator[chat_completion_chunk.ChoiceDeltaToolCall, None]:
    content_json = _load_content_json(content)
    for i, briton_fn in enumerate(content_json):
        if not (isinstance(briton_fn, dict) and "name" in briton_fn and "parameters" in briton_fn):
            logger.error(f"Generated tool calls {content_json} are not valid")
            continue

        name_delta = chat_completion_chunk.ChoiceDeltaToolCall(
            index=i,
            id=tool_call_id_fn(),
            function=chat_completion_chunk.ChoiceDeltaToolCallFunction(name=briton_fn["name"]),
            type="function",
        )
        yield name_delta

        args_str = json.dumps(briton_fn["parameters"])
        async for chunk in _chunk_args(args_str):
            args_delta = chat_completion_chunk.ChoiceDeltaToolCall(
                index=i,
                function=chat_completion_chunk.ChoiceDeltaToolCallFunction(arguments=chunk),
                type="function",
            )
            yield args_delta


def _create_completion_chunk(
    id: str,
    created: int,
    index: int,
    model: str,
    is_chat_completion: bool,
    content: Optional[str] = None,
    role: Optional[Literal["system", "user", "assistant", "tool"]] = None,
    finish_reason: Optional[
        Literal["stop", "length", "tool_calls", "content_filter", "function_call"]
    ] = None,
    tool_calls: Optional[List[chat_completion_chunk.ChoiceDeltaToolCall]] = None,
) -> ChatCompletionChunk | Completion:
    if is_chat_completion:
        delta = chat_completion_chunk.ChoiceDelta(content=content, role=role, tool_calls=tool_calls)
        choice = chat_completion_chunk.Choice(index=index, delta=delta, finish_reason=finish_reason)
        return ChatCompletionChunk(
            id=id,
            choices=[choice],
            created=created,
            model=model,
            object="chat.completion.chunk",
        )
    else:
        if content is None:
            content = ""
        if finish_reason is None:
            finish_reason = "length"
        choice = CompletionChoice(index=index, text=content, finish_reason=finish_reason)
        return Completion(
            id=id,
            choices=[choice],
            created=created,
            model=model,
            object="text_completion",
        )


async def _create_completion_chunks(
    created: int,
    req_id: str,
    index: int,
    model: str,
    input_text: AsyncGenerator[str, None],
    eos_token: Optional[str],
    tool_token: Optional[str],
    tool_call_id_fn: Optional[Callable[[], str]],
    stop_words: Optional[List[str]],
    skip_special_tokens: Optional[List[str]],
    is_chat_completion: bool,
) -> AsyncGenerator[ChatCompletionChunk | Completion, None]:
    start_chunk = _create_completion_chunk(
        id=req_id,
        created=created,
        index=index,
        model=model,
        content="",
        role="assistant",
        is_chat_completion=is_chat_completion,
    )
    is_first_iter = True
    delta = None
    async for delta in input_text:
        if is_first_iter:
            if tool_token is not None and delta.startswith(tool_token):
                break
            is_first_iter = False
            yield start_chunk

        content = remove_suffix_from_text(
            text=delta,
            eos_token=eos_token,
            stop_words=stop_words,
            skip_special_tokens=skip_special_tokens,
        )
        if len(content) == 0:
            continue  # Don't send empty chunks
        yield _create_completion_chunk(
            id=req_id,
            created=created,
            index=index,
            model=model,
            content=content,
            is_chat_completion=is_chat_completion,
        )

    if (
        is_first_iter
        and delta is not None
        and tool_token is not None
        and tool_call_id_fn is not None
        and delta.startswith(tool_token)
    ):
        full_text = delta.removeprefix(tool_token)
        async for delta in input_text:
            full_text += delta

        tool_calls = _create_tool_call_deltas(
            remove_suffix_from_text(
                text=full_text,
                eos_token=eos_token,
                stop_words=stop_words,
                skip_special_tokens=skip_special_tokens,
            ),
            tool_call_id_fn,
        )
        yield start_chunk
        async for tool_call_chunk in tool_calls:
            yield _create_completion_chunk(
                id=req_id,
                created=created,
                index=index,
                model=model,
                tool_calls=[tool_call_chunk],
                is_chat_completion=is_chat_completion,
            )
        finish_reason = "tool_calls"
    else:
        finish_reason = (
            _finish_reason_from_text(text=delta, eos_token=eos_token, stop_words=stop_words)
            if delta is not None
            else "length"
        )

    yield _create_completion_chunk(
        id=req_id,
        created=created,
        index=index,
        model=model,
        finish_reason=finish_reason,
        is_chat_completion=is_chat_completion,
    )


async def create_completion_chunks(
    req_id: str,
    model: str,
    sequences: List[AsyncGenerator[str, None]],
    eos_token: Optional[str] = None,
    tool_token: Optional[str] = None,
    tool_call_id_fn: Optional[Callable[[], str]] = None,
    prompt_tokens: Optional[int] = None,
    completion_tokens_fn: Optional[Callable[[], int]] = None,
    stop_words: Optional[List[str]] = None,
    skip_special_tokens: Optional[List[str]] = None,
    is_chat_completion: bool = False,
) -> AsyncGenerator[str, None]:
    created = int(time.time())

    chunk_generators = [
        _create_completion_chunks(
            created=created,
            req_id=req_id,
            index=i,
            model=model,
            input_text=input_text_gen,
            eos_token=eos_token,
            tool_token=tool_token,
            tool_call_id_fn=tool_call_id_fn,
            stop_words=stop_words,
            skip_special_tokens=skip_special_tokens,
            is_chat_completion=is_chat_completion,
        )
        for i, input_text_gen in enumerate(sequences)
    ]
    async for chunk in interleave_generators(*chunk_generators):
        yield _make_sse_chunk(chunk)

    if prompt_tokens is not None and completion_tokens_fn is not None:
        completion_tokens = completion_tokens_fn()
        usage = CompletionUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )
        if is_chat_completion:
            usage_chunk = ChatCompletionChunk(
                id=req_id,
                choices=[],
                created=created,
                model=model,
                object="chat.completion.chunk",
                usage=usage,
            )
        else:
            usage_chunk = Completion(
                id=req_id,
                choices=[],
                created=created,
                model=model,
                object="text_completion",
                usage=usage,
            )
        yield _make_sse_chunk(usage_chunk)

    yield "data: [DONE]\n\n"



from typing import Optional, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def common_prefix_length(list1: list, list2: list):
    i = 0
    while i < len(list1) and i < len(list2) and list1[i] == list2[i]:
        i += 1
    return i


def or_false(value: Optional[bool]) -> bool:
    return or_default(value, False)


def or_true(value: Optional[bool]) -> bool:
    return or_default(value, True)


def or_default(value: Optional[T], default_value: T) -> T:
    return value if value is not None else default_value


def optional_if(value: T, condition: bool) -> Optional[T]:
    return value if condition else None
